/* IRCfs - IRC FileServ for *nix.
 * Copyright (C) 2002  Nick 'Zaf' Clifford <zaf@nrc.co.nz>
 * For licensing details, refer to the LICENSE file in the source 
 * code directory.
 */
#include "str_hash.h"
#include "runtime.h"
/* TODO, make it a hash table, not a linked list! */
struct str_hash_node {
	char *name;
	void *data;
	struct str_hash_node *next, *prev;
	
};

struct str_hash_table {
	POOL pool;
	struct str_hash_node *head,*tail;
};

static struct str_hash_node *str_hash_get_node(struct str_hash_table *ht, 
		const char *name);
static struct str_hash_node *str_hash_get_node_n(struct str_hash_table *ht, 
		int num);

struct str_hash_table *str_hash_create(POOL parent)
{
	return str_hash_create_n(parent,10);
}

struct str_hash_table *str_hash_create_n(POOL parent, size_t buckets)
{
	struct str_hash_table *ht;
	POOL p = pool_new(parent);
	ht = XMALLOC(p,struct str_hash_table);
	ht->pool = p;
	return ht;
}
/* destroys a hash table.
 * free's any memory that *IT* allocated. eg
 * contents of nodes are *NOT* destroyed.
 * To do a total clean up, call str_hash_foreach, and free each
 * node yourself, then call str_hash_destroy
 */
int str_hash_destroy(struct str_hash_table *ht)
{
	ASSERT(ht);
	pool_destroy(ht->pool);
	return 0;
}

int str_hash_resize(struct str_hash_table *ht, size_t buckets)
{
	ASSERT(ht);
	return -1;
}

int str_hash_count(struct str_hash_table *ht)
{
	struct str_hash_node *n;
	int count=0;
	ASSERT(ht);
	for(n=ht->head; n != NULL; n=n->next) {
		count++;
	}
	return count;
}

int str_hash_foreach(struct str_hash_table *ht, str_hash_callback_t callback,
		void *appdata)
{
	struct str_hash_node *n;
	ASSERT(ht);
	ASSERT(callback);
	for(n=ht->head; n != NULL; n=n->next) {
		if (callback(ht, n->name, n->data,appdata) == -1)
			return -1;
	}
	return 0;
}

int str_hash_set(struct str_hash_table *ht, const char *name, void *contents,
		void **old)
{
	struct str_hash_node *n;
	ASSERT(ht);
	ASSERT(name);
	n = str_hash_get_node(ht,name);
	if (n != NULL) {
		if (old)
			*old = n->data;
		n->data = contents;
		return 0;
	}
	n = XMALLOC(ht->pool, struct str_hash_node);
	n->prev = ht->tail;
	if (ht->tail)
		ht->tail->next = n;
	if (!ht->head)
		ht->head = n;
	ht->tail = n;
	n->name = XSTRDUP(ht->pool, name);
	n->data = contents;
	return 0;


}
int str_hash_add(struct str_hash_table *ht, const char *name, void *contents)
{
	struct str_hash_node *n;
	ASSERT(ht);
	ASSERT(name);
	n = str_hash_get_node(ht,name);
	if (n != NULL) {
		return -1;
	}
	n = XMALLOC(ht->pool, struct str_hash_node);
	n->prev = ht->tail;
	if (ht->tail)
		ht->tail->next = n;
	if (!ht->head)
		ht->head = n;
	ht->tail = n;

	n->name = XSTRDUP(ht->pool, name);
	n->data = contents;
	return 0;
};
void *str_hash_del(struct str_hash_table *ht, const char *name)
{
	struct str_hash_node *n;
	void *data;
	ASSERT(ht);
	ASSERT(name);

	n = str_hash_get_node(ht,name);
	if (!n)
		return NULL;
	if (n->prev)
		n->prev->next = n->next;
	else
		ht->head = n->next;
	if (n->next)
		n->next->prev = n->prev;
	else
		ht->tail = n->prev;
	data = n->data;
	XFREE(n->name);
	XFREE(n);
	return data;
}
void *str_hash_get_n(struct str_hash_table *ht, int num)
{
	struct str_hash_node *n;
	ASSERT(ht);
	ASSERT(num >= 0);
	n=str_hash_get_node_n(ht,num);
	if (n)
		return n->data;
	return NULL;
}

void *str_hash_get(struct str_hash_table *ht, const char *name)
{
	struct str_hash_node *n;
	ASSERT(ht);
	ASSERT(name);
	n=str_hash_get_node(ht,name);
	if (n)
		return n->data;
	return NULL;
}
struct str_hash_node *str_hash_get_node_n(struct str_hash_table *ht, int num)
{
	struct str_hash_node *n;
	ASSERT(ht);
	ASSERT(num >= 0);
	if (num <= 0) 
		return NULL;
	for(n=ht->head; n != NULL; n=n->next) {
		if (--num <= 0) 
			return n;
	}
	return NULL;
}

struct str_hash_node *str_hash_get_node(struct str_hash_table *ht, const char *name)
{
	struct str_hash_node *n;
	ASSERT(ht);
	ASSERT(name);
	for(n=ht->head; n != NULL; n=n->next) {
		if (strcasecmp(n->name,name) == 0)
			return n;
	}
	return NULL;
}

